@@ -25,6 +25,18 @@ FunctionMinimum init2(const MnUserTransformation& tr, py::sequence par,
 py::tuple fmin_getstate(const FunctionMinimum&);
 FunctionMinimum fmin_setstate(py::tuple);
 
+bool hesse_failed_workaround(const FunctionMinimum& self) {
+  // Calling FunctionMinimum::HesseFailed is not reliable.
+  // It calls MinimumError::HesseFailed, but this is false if
+  // there were the failures MinimumError::MnReachedCallLimit
+  // or MinimumError::MnInvertFailed, which cannot be queried
+  // from FunctionMinimum.
+  //
+  // As a workaround, we return true if covariance exists, but
+  // is not accurate or made positive definite.
+  return self.HasCovariance() && !(self.HasPosDefCovar() || self.HasMadePosDefCovar());
+}
+
 void bind_functionminimum(py::module m) {
   py::class_<FunctionMinimum>(m, "FunctionMinimum")
 
@@ -42,7 +54,7 @@ void bind_functionminimum(py::module m) {
       .def_property_readonly("has_posdef_covar", &FunctionMinimum::HasPosDefCovar)
       .def_property_readonly("has_made_posdef_covar",
                              &FunctionMinimum::HasMadePosDefCovar)
-      .def_property_readonly("hesse_failed", &FunctionMinimum::HesseFailed)
+      .def_property_readonly("hesse_failed", hesse_failed_workaround)
       .def_property_readonly("has_covariance", &FunctionMinimum::HasCovariance)
       .def_property_readonly("is_above_max_edm", &FunctionMinimum::IsAboveMaxEdm)
       .def_property_readonly("has_reached_call_limit",
