@@ -1,4 +1,5 @@
-"""Jupyter-friendly Python interface for the Minuit2 library in C++.
+"""
+Jupyter-friendly Python interface for the Minuit2 library in C++.
 
 Basic usage example::
 
@@ -7,8 +8,6 @@ Basic usage example::
     def fcn(x, y, z):
         return (x - 2) ** 2 + (y - 3) ** 2 + (z - 4) ** 2
 
-    fcn.errordef = Minuit.LEAST_SQUARES
-
     m = Minuit(fcn, x=0, y=0, z=0)
     m.migrad()
     m.hesse()
@@ -21,15 +20,11 @@ Further information:
 * Code: https://github.com/scikit-hep/iminuit
 * Docs: https://iminuit.readthedocs.io
 """
+from iminuit.minuit import Minuit
+from iminuit.minimize import minimize
+from iminuit.util import describe
+from importlib import metadata
+
+__version__ = metadata.version("iminuit")
 
 __all__ = ["Minuit", "minimize", "describe", "__version__"]
-
-from .version import version as __version__
-
-try:
-    # make iminuit importable even if it is not installed yet for setup.cfg
-    from .minuit import Minuit
-    from .minimize import minimize
-    from .util import describe
-except ImportError:  # pragma: no cover
-    pass  # pragma: no cover
