@@ -22,8 +22,8 @@ def minimize(
     """
     Interface to MIGRAD using the ``scipy.optimize.minimize`` API.
 
-    This function provides the same interface as ``scipy.optimize.minimize``.
-    If you are familiar with the latter, this allows you to use Minuit with a quick start.
+    This function provides the same interface as ``scipy.optimize.minimize``. If you are
+    familiar with the latter, this allows you to use Minuit with a quick start.
     Eventually, you still may want to learn the interface of the :class:`Minuit` class,
     as it provides more functionality if you are interested in parameter uncertainties.
 
