@@ -14,9 +14,6 @@ def rosenbrock(x, y):
     return (1 - x) ** 2 + 100 * (y - x**2) ** 2
 
 
-rosenbrock.errordef = 1
-
-
 def rosenbrock_grad(x, y):
     """Gradient of Rosenbrock function."""
     return (-400 * x * (-(x**2) + y) + 2 * x - 2, -200 * x**2 + 200 * y)
@@ -35,9 +32,6 @@ def ackley(x, y):
     return term1 + term2 + 20 + e
 
 
-ackley.errordef = 1
-
-
 def beale(x, y):
     """
     Beale function. Minimum: f(3, 0.5) = 0.
@@ -50,9 +44,6 @@ def beale(x, y):
     return term1 * term1 + term2 * term2 + term3 * term3
 
 
-beale.errordef = 1
-
-
 def matyas(x, y):
     """
     Matyas function. Minimum: f(0, 0) = 0.
@@ -62,9 +53,6 @@ def matyas(x, y):
     return 0.26 * (x**2 + y**2) - 0.48 * x * y
 
 
-matyas.errordef = 1
-
-
 def sphere_np(x):
     """
     Sphere function for variable number of arguments. Minimum: f(0, ..., 0) = 0.
@@ -74,6 +62,3 @@ def sphere_np(x):
     import numpy as np
 
     return np.sum(x**2)
-
-
-sphere_np.errordef = 1
