@@ -1,3 +1,4 @@
+from __future__ import annotations
 import pytest
 import numpy as np
 from numpy.testing import assert_allclose, assert_equal
@@ -16,27 +17,10 @@ from iminuit.cost import (
     _soft_l1_loss,
     PerformanceWarning,
 )
+from iminuit.util import describe
+from iminuit.typing import Annotated, Gt, Lt
 from typing import Sequence
 import pickle
-from sys import version_info as pyver
-
-try:
-    # pytest.importorskip does not work for scipy.stats;
-    # for some reason, import scipy or import scipy.stats succeeds
-    # even if scipy is not installed
-    from scipy.stats import norm, truncexpon, multivariate_normal
-
-    scipy_available = True
-except ImportError:
-    scipy_available = False
-
-
-try:
-    from matplotlib import pyplot  # noqa
-
-    matplotlib_available = True
-except ImportError:
-    matplotlib_available = False
 
 
 def norm_logpdf(x, mu, sigma):
@@ -48,13 +32,22 @@ def norm_pdf(x, mu, sigma):
     return np.exp(norm_logpdf(x, mu, sigma))
 
 
+def norm_cdf(x, mu, sigma):
+    from math import erf
+
+    z = (x - mu) / (np.sqrt(2) * sigma)
+
+    return (1 + np.vectorize(erf)(z)) * 0.5
+
+
 def mvnorm(mux, muy, sx, sy, rho):
+    stats = pytest.importorskip("scipy.stats")
     C = np.empty((2, 2))
     C[0, 0] = sx**2
     C[0, 1] = C[1, 0] = sx * sy * rho
     C[1, 1] = sy**2
     m = [mux, muy]
-    return multivariate_normal(m, C)
+    return stats.multivariate_normal(m, C)
 
 
 def expon_cdf(x, a):
@@ -87,29 +80,69 @@ def pdf(x, mu, sigma):
 
 
 def cdf(x, mu, sigma):
-    return norm(mu, sigma).cdf(x)
+    return norm_cdf(x, mu, sigma)
 
 
 def scaled_cdf(x, n, mu, sigma):
-    return n * norm(mu, sigma).cdf(x)
+    return n * norm_cdf(x, mu, sigma)
 
 
 def line(x, a, b):
     return a + b * x
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_norm_logpdf():
+    norm = pytest.importorskip("scipy.stats").norm
     x = np.linspace(-3, 3)
     assert_allclose(norm_logpdf(x, 3, 2), norm.logpdf(x, 3, 2))
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_norm_pdf():
+    norm = pytest.importorskip("scipy.stats").norm
     x = np.linspace(-3, 3)
     assert_allclose(norm_pdf(x, 3, 2), norm.pdf(x, 3, 2))
 
 
+def test_norm_cdf():
+    norm = pytest.importorskip("scipy.stats").norm
+    x = np.linspace(-3, 3)
+    assert_allclose(norm_cdf(x, 3, 2), norm.cdf(x, 3, 2))
+
+
+def test_describe():
+    c = Constant(2.5)
+    assert describe(c, annotations=True) == {}
+
+    c = NormalConstraint("foo", 1.5, 0.1)
+    assert describe(c, annotations=True) == {"foo": None}
+
+    def model(
+        x, foo: Annotated[float, Gt(1), Lt(2)], bar: float, baz: Annotated[float, 0:]
+    ):
+        return x
+
+    assert describe(model, annotations=True) == {
+        "x": None,
+        "foo": (1, 2),
+        "bar": None,
+        "baz": (0, np.inf),
+    }
+
+    c = UnbinnedNLL([], model)
+    assert describe(c, annotations=True) == {
+        "foo": (1, 2),
+        "bar": None,
+        "baz": (0, np.inf),
+    }
+
+    c = BinnedNLL([], [1], model)
+    assert describe(c, annotations=True) == {
+        "foo": (1, 2),
+        "bar": None,
+        "baz": (0, np.inf),
+    }
+
+
 def test_Constant():
     c = Constant(2.5)
     assert c.value == 2.5
@@ -133,7 +166,6 @@ def test_UnbinnedNLL(unbinned, verbose, model):
     assert_equal(m.fmin.reduced_chi2, np.nan)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_UnbinnedNLL_2D():
     def model(x_y, mux, muy, sx, sy, rho):
         return mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)
@@ -155,10 +187,10 @@ def test_UnbinnedNLL_mask():
     c = UnbinnedNLL([1, np.nan, 2], lambda x, a: x + a)
     assert c.mask is None
 
-    assert np.isnan(c(0)) == True
+    assert np.isnan(c(0))
     c.mask = np.arange(3) != 1
     assert_equal(c.mask, (True, False, True))
-    assert np.isnan(c(0)) == False
+    assert not np.isnan(c(0))
 
 
 @pytest.mark.parametrize("log", (False, True))
@@ -188,21 +220,32 @@ def test_UnbinnedNLL_properties(log):
     assert c.verbose == 1
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 @pytest.mark.parametrize("log", (False, True))
 def test_UnbinnedNLL_visualize(log):
+    pytest.importorskip("matplotlib")
     c = UnbinnedNLL([1, 2], norm_logpdf if log else norm_pdf, log=log)
-    c.visualize((1, 2))  # auto-sampling
-    c.visualize((1, 2), model_points=10)  # linear spacing
-
-    # trigger log-spacing
-    c = UnbinnedNLL([1, 1000], norm_logpdf if log else norm_pdf, log=log)
+    # auto-sampling
+    c.visualize((1, 2))
+    # linear spacing
     c.visualize((1, 2), model_points=10)
+    # linear spacing and different binning
+    c.visualize((1, 2), model_points=10, bins=20)
+    # trigger log-spacing
+    c = UnbinnedNLL([1, 100, 1000], norm_logpdf if log else norm_pdf, log=log)
+    c.visualize((1, 2), model_points=10)
+    c.visualize((1, 2), model_points=10, bins=20)
+    # manual spacing
+    c.visualize((1, 2), model_points=np.linspace(1, 1000))
+
+    with pytest.warns(
+        np.VisibleDeprecationWarning, match="keyword 'nbins' is deprecated"
+    ):
+        c.visualize((1, 2), nbins=20)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_UnbinnedNLL_visualize_2D():
+    pytest.importorskip("matplotlib")
+
     def model(x_y, mux, muy, sx, sy, rho):
         return mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)
 
@@ -222,6 +265,20 @@ def test_UnbinnedNLL_pickle():
     assert_equal(c.data, c2.data)
 
 
+def test_UnbinnedNLL_annotated():
+    def model(
+        x: np.ndarray,
+        mu: float,
+        sigma: Annotated[float, 0 : np.inf],
+    ) -> np.ndarray:
+        return norm_pdf(x, mu, sigma)
+
+    c = UnbinnedNLL([], model)
+    m = Minuit(c, mu=1, sigma=1.5)
+    assert m.limits[0] == (-np.inf, np.inf)
+    assert m.limits[1] == (0, np.inf)
+
+
 @pytest.mark.parametrize("verbose", (0, 1))
 @pytest.mark.parametrize("model", (logpdf, pdf))
 def test_ExtendedUnbinnedNLL(unbinned, verbose, model):
@@ -247,7 +304,6 @@ def test_ExtendedUnbinnedNLL(unbinned, verbose, model):
     assert_equal(m.fmin.reduced_chi2, np.nan)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_ExtendedUnbinnedNLL_2D():
     def model(x_y, n, mux, muy, sx, sy, rho):
         return n * 1000, n * 1000 * mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)
@@ -270,9 +326,9 @@ def test_ExtendedUnbinnedNLL_mask():
     c = ExtendedUnbinnedNLL([1, np.nan, 2], lambda x, a: (1, x + a))
     assert c.ndata == np.inf
 
-    assert np.isnan(c(0)) == True
+    assert np.isnan(c(0))
     c.mask = np.arange(3) != 1
-    assert np.isnan(c(0)) == False
+    assert not np.isnan(c(0))
     assert c.ndata == np.inf
 
 
@@ -303,8 +359,9 @@ def test_ExtendedUnbinnedNLL_properties(log):
 
 
 @pytest.mark.parametrize("log", (False, True))
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_ExtendedUnbinnedNLL_visualize(log):
+    pytest.importorskip("matplotlib")
+
     def log_model(x, s, mu, sigma):
         return s, np.log(s) + norm_logpdf(x, mu, sigma)
 
@@ -320,9 +377,9 @@ def test_ExtendedUnbinnedNLL_visualize(log):
     c.visualize((1, 2, 3))
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_ExtendedUnbinnedNLL_visualize_2D():
+    pytest.importorskip("matplotlib")
+
     def model(x_y, n, mux, muy, sx, sy, rho):
         return n * 100, n * 100 * mvnorm(mux, muy, sx, sy, rho).pdf(x_y.T)
 
@@ -342,7 +399,6 @@ def test_ExtendedUnbinnedNLL_pickle():
     assert_equal(c.data, c2.data)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 @pytest.mark.parametrize("verbose", (0, 1))
 def test_BinnedNLL(binned, verbose):
     mle, nx, xe = binned
@@ -361,8 +417,19 @@ def test_BinnedNLL(binned, verbose):
     assert_allclose(m.fmin.reduced_chi2, 1, atol=0.15)
 
 
+def test_BinnedNLL_pulls(binned):
+    mle, nx, xe = binned
+    cost = BinnedNLL(nx, xe, cdf)
+    m = Minuit(cost, mu=0, sigma=1)
+    m.limits["sigma"] = (0, None)
+    m.migrad()
+    pulls = cost.pulls(m.values)
+    assert np.nanmean(pulls) == pytest.approx(0, abs=0.05)
+    assert np.nanvar(pulls) == pytest.approx(1, abs=0.2)
+
+
 def test_BinnedNLL_weighted():
-    xe = np.array([0, 1, 10])
+    xe = np.array([0, 0.2, 0.4, 0.8, 1.5, 10])
     p = np.diff(expon_cdf(xe, 1))
     n = p * 1000
     c = BinnedNLL(n, xe, expon_cdf)
@@ -372,6 +439,8 @@ def test_BinnedNLL_weighted():
     m1.migrad()
     assert m1.values[0] == pytest.approx(1, rel=1e-2)
 
+    # variance * 4 = (sigma * 2)^2, constructed so that
+    # fitted parameter has twice the uncertainty
     w = np.transpose((n, 4 * n))
     c = BinnedNLL(w, xe, expon_cdf)
     assert_equal(c.data, w)
@@ -381,6 +450,25 @@ def test_BinnedNLL_weighted():
     assert m2.errors[0] == pytest.approx(2 * m1.errors[0], rel=1e-2)
 
 
+def test_ExtendedBinnedNLL_weighted_pulls():
+    rng = np.random.default_rng(1)
+    xe = np.linspace(0, 5, 500)
+    p = np.diff(expon_cdf(xe, 2))
+    m = p * 100000
+    n = rng.poisson(m * 4) / 4
+    nvar = m * 4 / 4**2
+    w = np.transpose((n, nvar))
+    c = ExtendedBinnedNLL(w, xe, lambda x, n, s: n * expon_cdf(x, s))
+    m = Minuit(c, np.sum(n), 2)
+    m.limits["n"] = (0.1, None)
+    m.limits["s"] = (0.1, None)
+    m.migrad()
+    assert m.valid
+    pulls = c.pulls(m.values)
+    assert np.nanmean(pulls) == pytest.approx(0, abs=0.1)
+    assert np.nanvar(pulls) == pytest.approx(1, abs=0.2)
+
+
 def test_BinnedNLL_bad_input_1():
     with pytest.raises(ValueError):
         BinnedNLL([1], [1], lambda x, a: 0)
@@ -419,7 +507,6 @@ def test_BinnedNLL_bad_input_6():
         BinnedNLL(1, 2, lambda x, a: 0)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_BinnedNLL_2D():
     truth = (0.1, 0.2, 0.3, 0.4, 0.5)
     x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T
@@ -445,7 +532,6 @@ def test_BinnedNLL_2D():
     assert cost(*m.values) > m.fval
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_BinnedNLL_2D_with_zero_bins():
     truth = (0.1, 0.2, 0.3, 0.4, 0.5)
     x, y = mvnorm(*truth).rvs(size=1000, random_state=1).T
@@ -466,7 +552,6 @@ def test_BinnedNLL_2D_with_zero_bins():
 
 
 def test_BinnedNLL_mask():
-
     c = BinnedNLL([5, 1000, 1], [0, 1, 2, 3], expon_cdf)
     assert c.ndata == 3
 
@@ -492,8 +577,9 @@ def test_BinnedNLL_properties():
         c.n = [1, 2]
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_BinnedNLL_visualize():
+    pytest.importorskip("matplotlib")
+
     c = BinnedNLL([1, 2], [1, 2, 3], expon_cdf)
 
     c.visualize((1,))
@@ -502,9 +588,9 @@ def test_BinnedNLL_visualize():
     c.visualize((1,))
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_BinnedNLL_visualize_2D():
+    pytest.importorskip("matplotlib")
+
     truth = (0.1, 0.2, 0.3, 0.4, 0.5)
     x, y = mvnorm(*truth).rvs(size=10, random_state=1).T
     w, xe, ye = np.histogram2d(x, y, bins=(50, 100), range=((-5, 5), (-5, 5)))
@@ -514,8 +600,7 @@ def test_BinnedNLL_visualize_2D():
 
     c = BinnedNLL(w, (xe, ye), model)
 
-    with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
-        c.visualize(truth)
+    c.visualize(truth)
 
 
 def test_BinnedNLL_pickle():
@@ -525,7 +610,6 @@ def test_BinnedNLL_pickle():
     assert_equal(c.data, c2.data)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 @pytest.mark.parametrize("verbose", (0, 1))
 def test_ExtendedBinnedNLL(binned, verbose):
     mle, nx, xe = binned
@@ -565,7 +649,6 @@ def test_ExtendedBinnedNLL_bad_input():
         ExtendedBinnedNLL([1], [1], lambda x, a: 0)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_ExtendedBinnedNLL_2D():
     truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
     x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 1000), random_state=1).T
@@ -585,8 +668,9 @@ def test_ExtendedBinnedNLL_2D():
     assert_allclose(m.values, truth, atol=0.1)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 def test_ExtendedBinnedNLL_3D():
+    norm = pytest.importorskip("scipy.stats").norm
+
     truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7)
     n = int(truth[0] * 10000)
     x, y = mvnorm(*truth[1:-2]).rvs(size=n).T
@@ -631,8 +715,9 @@ def test_ExtendedBinnedNLL_properties():
     assert c.scaled_cdf is cdf
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_ExtendedBinnedNLL_visualize():
+    pytest.importorskip("matplotlib")
+
     def model(x, s, slope):
         return s * expon_cdf(x, slope)
 
@@ -640,9 +725,9 @@ def test_ExtendedBinnedNLL_visualize():
     c.visualize((1, 2))
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_ExtendedBinnedNLL_visualize_2D():
+    pytest.importorskip("matplotlib")
+
     truth = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
     x, y = mvnorm(*truth[1:]).rvs(size=int(truth[0] * 1000), random_state=1).T
 
@@ -653,8 +738,7 @@ def test_ExtendedBinnedNLL_visualize_2D():
 
     c = ExtendedBinnedNLL(w, (xe, ye), model)
 
-    with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
-        c.visualize(truth)
+    c.visualize(truth)
 
 
 def test_ExtendedBinnedNLL_pickle():
@@ -718,20 +802,23 @@ def test_LeastSquares_2D():
 
 
 def test_LeastSquares_bad_input():
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match="shape mismatch"):
         LeastSquares([1, 2], [], [1], lambda x, a: 0)
 
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match="shape mismatch"):
         LeastSquares([1, 2], [3, 4, 5], [1], lambda x, a: 0)
 
-    with pytest.raises(ValueError):
+    with pytest.raises(ValueError, match="unknown loss"):
         LeastSquares([1], [1], [1], lambda x, a: 0, loss="foo")
 
+    with pytest.raises(ValueError, match="loss must be str or LossFunction"):
+        LeastSquares([1], [1], [1], lambda x, a: 0, loss=[1, 2, 3])
+
 
 def test_LeastSquares_mask():
     c = LeastSquares([1, 2, 3], [3, np.nan, 4], [1, 1, 1], lambda x, a: x + a)
     assert c.ndata == 3
-    assert np.isnan(c(0)) == True
+    assert np.isnan(c(0))
 
     m = Minuit(c, 1)
     assert m.ndof == 2
@@ -739,7 +826,7 @@ def test_LeastSquares_mask():
     assert not m.valid
 
     c.mask = np.arange(3) != 1
-    assert np.isnan(c(0)) == False
+    assert not np.isnan(c(0))
     assert c.ndata == 2
 
     assert m.ndof == 1
@@ -779,20 +866,25 @@ def test_LeastSquares_properties():
         c.yerror = [1, 2]
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_LeastSquares_visualize():
+    pytest.importorskip("matplotlib")
+
     c = LeastSquares([1, 2], [2, 3], 0.1, line)
 
-    c.visualize((1, 2))  # auto-sampling
-    c.visualize((1, 2), model_points=10)  # linear spacing
-
-    # trigger use of log-spacing
-    c = LeastSquares([1, 2000], [2, 3], 0.1, line)
+    # auto-sampling
+    c.visualize((1, 2))
+    # linear spacing
     c.visualize((1, 2), model_points=10)
+    # trigger use of log-spacing
+    c = LeastSquares([1, 10, 100], [2, 3, 4], 0.1, line)
+    c.visualize((1, 2), model_points=10)
+    # manual spacing
+    c.visualize((1, 2), model_points=np.linspace(1, 100))
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_LeastSquares_visualize_2D():
+    pytest.importorskip("matplotlib")
+
     c = LeastSquares([[1, 2]], [[2, 3]], 0.1, line)
 
     with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
@@ -800,7 +892,6 @@ def test_LeastSquares_visualize_2D():
 
 
 def test_LeastSquares_pickle():
-
     c = LeastSquares([1, 2], [2, 3], 0.1, line)
 
     b = pickle.dumps(c)
@@ -809,6 +900,16 @@ def test_LeastSquares_pickle():
     assert c.model is c2.model
 
 
+def test_LeastSquares_pulls():
+    c = LeastSquares([1, 2], [2, 3], 0.1, line)
+    # pulls are zero for parameters 1, 1
+    assert_equal(c.pulls((1, 1)), [0, 0])
+    # correct slope but line offset by 1 with error 0.1 produces pulls [10, 10]
+    assert_equal(c.pulls((0, 1)), [10, 10])
+    c.mask = [True, False]
+    assert_equal(c.pulls((0, 1)), [10, np.nan])
+
+
 def test_CostSum_1():
     def model1(x, a):
         return a + x
@@ -820,20 +921,20 @@ def test_CostSum_1():
         return c
 
     lsq1 = LeastSquares(1, 2, 3, model1)
-    assert lsq1.func_code.co_varnames == ("a",)
+    assert describe(lsq1) == ["a"]
 
     lsq2 = LeastSquares(1, 3, 4, model2)
-    assert lsq2.func_code.co_varnames == ("b", "a")
+    assert describe(lsq2) == ["b", "a"]
 
     lsq3 = LeastSquares(1, 1, 1, model3)
-    assert lsq3.func_code.co_varnames == ("c",)
+    assert describe(lsq3) == ["c"]
 
     lsq12 = lsq1 + lsq2
     assert lsq12._items == [lsq1, lsq2]
     assert isinstance(lsq12, CostSum)
     assert isinstance(lsq1, LeastSquares)
     assert isinstance(lsq2, LeastSquares)
-    assert lsq12.func_code.co_varnames == ("a", "b")
+    assert describe(lsq12) == ["a", "b"]
     assert lsq12.ndata == 2
 
     assert lsq12(1, 2) == lsq1(1) + lsq2(2, 1)
@@ -847,22 +948,22 @@ def test_CostSum_1():
 
     lsq121 = lsq12 + lsq1
     assert lsq121._items == [lsq1, lsq2, lsq1]
-    assert lsq121.func_code.co_varnames == ("a", "b")
+    assert describe(lsq121) == ["a", "b"]
     assert lsq121.ndata == 3
 
     lsq312 = lsq3 + lsq12
     assert lsq312._items == [lsq3, lsq1, lsq2]
-    assert lsq312.func_code.co_varnames == ("c", "a", "b")
+    assert describe(lsq312) == ["c", "a", "b"]
     assert lsq312.ndata == 3
 
     lsq31212 = lsq312 + lsq12
     assert lsq31212._items == [lsq3, lsq1, lsq2, lsq1, lsq2]
-    assert lsq31212.func_code.co_varnames == ("c", "a", "b")
+    assert describe(lsq31212) == ["c", "a", "b"]
     assert lsq31212.ndata == 5
 
     lsq31212 += lsq1
     assert lsq31212._items == [lsq3, lsq1, lsq2, lsq1, lsq2, lsq1]
-    assert lsq31212.func_code.co_varnames == ("c", "a", "b")
+    assert describe(lsq31212) == ["c", "a", "b"]
     assert lsq31212.ndata == 6
 
 
@@ -894,8 +995,8 @@ def test_CostSum_3():
     assert cs((1, 1)) == lsq((1, 1)) + con((1, 1)) + 1.5
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy is needed")
 def test_CostSum_4():
+    pytest.importorskip("scipy.special")
 
     t = Template([1, 2], [1, 2, 3], [[1, 1], [0, 1]], method="asy")
     assert t.errordef == Minuit.LIKELIHOOD
@@ -912,8 +1013,9 @@ def test_CostSum_4():
     assert_allclose(m1.errors, m2.errors)
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_CostSum_visualize():
+    pytest.importorskip("matplotlib")
+
     lsq = LeastSquares([1, 2, 3], [3, 4, 5], 1, line)
     con = NormalConstraint(("a", "b"), (1, 1), (1, 1))
     c = lsq + con + 1
@@ -926,8 +1028,8 @@ def test_NormalConstraint_1():
 
     lsq1 = LeastSquares(0, 1, 1, model)
     lsq2 = lsq1 + NormalConstraint("a", 1, 0.1)
-    assert lsq1.func_code.co_varnames == ("a",)
-    assert lsq2.func_code.co_varnames == ("a",)
+    assert describe(lsq1) == ["a"]
+    assert describe(lsq2) == ["a"]
     assert lsq1.ndata == 1
     assert lsq2.ndata == 2
 
@@ -950,9 +1052,9 @@ def test_NormalConstraint_2():
     rho = 0.5
     cov = ((sa**2, rho * sa * sb), (rho * sa * sb, sb**2))
     lsq3 = lsq1 + NormalConstraint(("a", "b"), (1, 2), cov)
-    assert lsq1.func_code.co_varnames == ("a", "b")
-    assert lsq2.func_code.co_varnames == ("a", "b")
-    assert lsq3.func_code.co_varnames == ("a", "b")
+    assert describe(lsq1) == ["a", "b"]
+    assert describe(lsq2) == ["a", "b"]
+    assert describe(lsq3) == ["a", "b"]
     assert lsq1.ndata == 2
     assert lsq2.ndata == 4
 
@@ -983,8 +1085,9 @@ def test_NormalConstraint_properties():
     assert_equal(nc.covariance, (1, 2))
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_NormalConstraint_visualize():
+    pytest.importorskip("matplotlib")
+
     c = NormalConstraint(("a", "b"), (1, 2), (3, 4))
     c.visualize((1, 2))
 
@@ -996,7 +1099,7 @@ def test_NormalConstraint_pickle():
     c = NormalConstraint(("a", "b"), (1, 2), (3, 4))
     b = pickle.dumps(c)
     c2 = pickle.loads(b)
-    assert c.func_code == c2.func_code
+    assert describe(c) == describe(c2) and describe(c) == ["a", "b"]
     assert_equal(c.value, c2.value)
     assert_equal(c.covariance, c2.covariance)
 
@@ -1022,7 +1125,7 @@ def test_multinominal_chi2():
 
     assert multinominal_chi2(zero, zero) == 0
     assert multinominal_chi2(zero, one) == 0
-    assert multinominal_chi2(one, zero) == pytest.approx(1487, abs=1)
+    assert multinominal_chi2(one, zero) == pytest.approx(1416, abs=1)
     n = np.array([(0.0, 0.0)])
     assert_allclose(multinominal_chi2(n, zero), 0)
     assert_allclose(multinominal_chi2(n, one), 0)
@@ -1091,14 +1194,17 @@ def test_update_data_with_mask(cls):
 
 @pytest.mark.parametrize("method", ("jsc", "asy", "da"))
 def test_Template(method):
-    if method == "asy" and not scipy_available:
-        pytest.skip(reason="scipy needed")
+    if method == "asy":
+        pytest.importorskip("scipy.special")
+
     xe = np.array([0, 1, 2, 3])
     t = np.array([[1, 1, 0], [0, 1, 3]])
     n = t[0] + t[1]
 
     c = Template(n, xe, t, method=method)
     m = Minuit(c, 1, 1)
+    assert m.limits[0] == (0, np.inf)
+    assert m.limits[1] == (0, np.inf)
     m.migrad()
     assert m.valid
     assert m.ndof == 1
@@ -1114,10 +1220,32 @@ def test_Template(method):
         assert_allclose(m.values, [2, 4], atol=1e-2)
 
 
+@pytest.mark.parametrize(
+    "template",
+    (
+        (np.array([1.0, 1.0, 0.0]), np.array([0.0, 1.0, 3.0])),
+        (
+            np.array([[1.0, 1.0], [1.0, 1.0], [0.0, 0.0]]),
+            np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]]),
+        ),
+    ),
+)
+def test_Template_does_not_modify_inputs(template):
+    from copy import deepcopy
+
+    xe = np.array([0, 1, 2, 3])
+    template_copy = deepcopy(template)
+    n = template[0] + template[1]
+
+    Template(n, xe, template, method="da")
+    assert_equal(template, template_copy)
+
+
 def generate(rng, nmc, truth, bins, tf=1, df=1):
+    stats = pytest.importorskip("scipy.stats")
     xe = np.linspace(0, 2, bins + 1)
-    b = np.diff(truncexpon(1, 0, 2).cdf(xe))
-    s = np.diff(norm(1, 0.1).cdf(xe))
+    b = np.diff(stats.truncexpon(1, 0, 2).cdf(xe))
+    s = np.diff(stats.norm(1, 0.1).cdf(xe))
     n = b * truth[0] + s * truth[1]
     t = b * nmc, s * nmc
     if rng is not None:
@@ -1130,13 +1258,10 @@ def generate(rng, nmc, truth, bins, tf=1, df=1):
     return n, xe, np.array(t)
 
 
-@pytest.mark.skipif(not scipy_available, reason="scipy.stats is needed")
 @pytest.mark.parametrize("method", ("jsc", "asy", "da"))
 @pytest.mark.parametrize("with_mask", (False, True))
 @pytest.mark.parametrize("weighted_data", (False, True))
 def test_Template_weighted(method, with_mask, weighted_data):
-    if method == "asy" and not scipy_available:
-        pytest.skip(reason="scipy needed")
     rng = np.random.default_rng(1)
     truth = 750, 250
     z = []
@@ -1181,9 +1306,13 @@ def test_Template_bad_input():
     with pytest.raises(ValueError, match="number of names"):
         Template([1], [1, 2], [[1]], name=("b", "s"))
 
+    with pytest.raises(ValueError, match="model_or_template"):
+        Template([1], [1, 2], [[1], None])
+
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_Template_visualize():
+    pytest.importorskip("matplotlib")
+
     xe = [0, 1, 2]
     n = [1, 2]
     t = [[1, 2], [5, 4]]
@@ -1196,16 +1325,15 @@ def test_Template_visualize():
     c.visualize((1, 2))
 
 
-@pytest.mark.skipif(not matplotlib_available, reason="matplotlib is needed")
 def test_Template_visualize_2D():
+    pytest.importorskip("matplotlib")
+
     xe = ([0, 1, 2], [0, 1, 2])
     n = [[1, 2], [3, 4]]
     t = [[[1, 2], [1, 2]], [[5, 4], [5, 4]]]
 
     c = Template(n, xe, t)
-
-    with pytest.raises(ValueError, match="not implemented for multi-dimensional"):
-        c.visualize((1, 2))
+    c.visualize((1, 2))
 
 
 def test_Template_pickle():
@@ -1220,7 +1348,91 @@ def test_Template_pickle():
     assert_equal(c.data, c2.data)
 
 
-@pytest.mark.skipif(pyver < (3, 7), reason="module getattr requires Python-3.7+")
+def test_Template_with_model():
+    n = np.array([3, 2, 3])
+    xe = np.array([0, 1, 2, 3])
+    t = np.array([0.1, 0, 1])
+
+    c = Template(n, xe, (t, scaled_cdf))
+    assert describe(c) == ["x0", "x1_n", "x1_mu", "x1_sigma"]
+
+    m = Minuit(c, 1, 1, 0.5, 1)
+    m.limits["x0", "x1_n"] = (0, None)
+    m.limits["x1_sigma"] = (0.1, None)
+    m.limits["x1_mu"] = (xe[0], xe[-1])
+    m.migrad()
+    assert m.valid
+
+    c2 = Template(n, xe, (t, scaled_cdf), name=("a", "b", "c", "d"))
+    assert describe(c2) == ["a", "b", "c", "d"]
+    m = Minuit(c2, 1, 1, 0.5, 1)
+    m.limits["d"] = (0.1, None)
+    m.migrad()
+    assert m.valid
+
+
+def test_Template_with_model_2D():
+    truth1 = (1.0, 0.1, 0.2, 0.3, 0.4, 0.5)
+    x1, y1 = mvnorm(*truth1[1:]).rvs(size=int(truth1[0] * 1000), random_state=1).T
+    truth2 = (1.0, 0.2, 0.1, 0.4, 0.3, 0.0)
+    x2, y2 = mvnorm(*truth2[1:]).rvs(size=int(truth2[0] * 1000), random_state=1).T
+
+    x = np.append(x1, x2)
+    y = np.append(y1, y2)
+    w, xe, ye = np.histogram2d(x, y, bins=(3, 5))
+
+    def model(xy, n, mux, muy, sx, sy, rho):
+        return n * 1000 * mvnorm(mux, muy, sx, sy, rho).cdf(np.transpose(xy))
+
+    x3, y3 = mvnorm(*truth2[1:]).rvs(size=int(truth2[0] * 10000), random_state=2).T
+    template = np.histogram2d(x3, y3, bins=(xe, ye))[0]
+
+    cost = Template(w, (xe, ye), (model, template))
+    assert cost.ndata == np.prod(w.shape)
+    m = Minuit(cost, *truth1, 1)
+    m.limits["x0_n", "x0_sx", "x0_sy"] = (0, None)
+    m.limits["x0_rho"] = (-1, 1)
+    m.migrad()
+    assert m.valid
+    assert_allclose(m.values, truth1 + (1e3,), rtol=0.1)
+
+
+def test_Template_with_only_models():
+    n = np.array([3, 2, 3])
+    xe = np.array([0, 1, 2, 3])
+
+    c = Template(n, xe, (lambda x, n, mu: n * expon_cdf(x, mu), scaled_cdf))
+    assert describe(c) == ["x0_n", "x0_mu", "x1_n", "x1_mu", "x1_sigma"]
+
+    m = Minuit(c, 1, 0.5, 1, 0.5, 1)
+    m.limits["x0_n", "x1_n", "x1_sigma"] = (0, None)
+    m.limits["x0_mu", "x1_mu"] = (xe[0], xe[-1])
+    m.migrad()
+    assert m.valid
+
+
+def test_Template_pulls():
+    rng = np.random.default_rng(1)
+    xe = np.linspace(0, 1, 200)
+    tmu = np.array(
+        [
+            np.diff(norm_cdf(xe, 0.5, 0.1)) * 1000,
+            np.diff(expon_cdf(xe, 1)) * 10000,
+        ]
+    )
+    truth = (1000, 2000)
+    mu = sum(truth[i] * (tmu[i] / np.sum(tmu[i])) for i in range(2))
+    n = rng.poisson(mu)
+    t = rng.poisson(tmu)
+    c = Template(n, xe, t)
+    m = Minuit(c, *truth)
+    m.migrad()
+    assert m.valid
+    pulls = c.pulls(m.values)
+    assert np.nanmean(pulls) == pytest.approx(0, abs=0.1)
+    assert np.nanvar(pulls) == pytest.approx(1, abs=0.1)
+
+
 def test_deprecated():
     from iminuit import cost
 
@@ -1243,3 +1455,43 @@ def test_deprecated_Template_method():
     with pytest.warns(np.VisibleDeprecationWarning):
         t = Template([1], [2, 3], [[1], [2]], method="hpd")
         t._impl is cost.template_chi2_da
+
+
+@pytest.mark.parametrize("cost", (BinnedNLL, ExtendedBinnedNLL))
+def test_binned_cost_with_model_shape_error_message_1D(cost):
+    n = [1, 2]
+    edges = [0.1, 0.2, 0.3]
+
+    # cdf should return same length as xe, not same length as n
+    def cdf(xe, a):
+        return xe[:-1]
+
+    c = cost(n, edges, cdf)
+    with pytest.raises(
+        ValueError,
+        match=(
+            r"Expected model to return an array of shape \(3,\), "
+            r"but it returns an array of shape \(2,\)"
+        ),
+    ):
+        c(1)
+
+
+@pytest.mark.parametrize("cost", (BinnedNLL, ExtendedBinnedNLL))
+def test_binned_cost_with_model_shape_error_message_2D(cost):
+    n = [[1, 2, 3], [4, 5, 6]]
+    edges = [0.5, 0.6, 0.7], [0.1, 0.2, 0.3, 0.4]
+
+    def cdf(xye, a):
+        # xye has shape (2, N), returned array should be (N,)
+        return np.ones(xye.shape[1] - 1)
+
+    c = cost(n, edges, cdf)
+    with pytest.raises(
+        ValueError,
+        match=(
+            r"Expected model to return an array of shape \(12,\), "
+            r"but it returns an array of shape \(11,\)"
+        ),
+    ):
+        c(1)
