@@ -7,6 +7,7 @@ import contextlib
 
 mpl = pytest.importorskip("matplotlib")
 plt = pytest.importorskip("matplotlib.pyplot")
+
 mpl.use("Agg")
 
 
@@ -32,40 +33,24 @@ def fig(request):
     plt.close()
 
 
-@pytest.mark.parametrize("arg", ("x", "y"))
+@pytest.mark.parametrize("arg", ("x", 1))
 def test_profile_1(fig, minuit, arg):
-    # plots with hesse errors
     minuit.draw_profile(arg)
     plt.ylim(0, 5)
 
 
-@pytest.mark.parametrize("arg", ("x", "y"))
-def test_profile_2(fig, minuit, arg):
-    # plots with minos errors
-    minuit.draw_profile(arg)
-    plt.ylim(0, 5)
-
-
-def test_profile_3(fig, minuit):
+def test_profile_2(fig, minuit):
     minuit.draw_profile("x", grid=np.linspace(0, 5))
 
 
-@pytest.mark.parametrize("arg", ("x", "y"))
+@pytest.mark.parametrize("arg", ("x", 1))
 def test_mnprofile_1(fig, minuit, arg):
     # plots with hesse errors
     minuit.draw_mnprofile(arg)
     plt.ylim(0, 5)
 
 
-@pytest.mark.parametrize("arg", ("x", "y"))
-def test_mnprofile_2(fig, minuit, arg):
-    # plots with minos errors
-    minuit.minos()
-    minuit.draw_mnprofile(arg)
-    plt.ylim(0, 5)
-
-
-def test_mnprofile_3(fig, minuit):
+def test_mnprofile_2(fig, minuit):
     minuit.minos()
     minuit.draw_mnprofile("x", grid=np.linspace(0, 5))
 
@@ -75,7 +60,8 @@ def test_mncontour_1(fig, minuit):
 
 
 def test_mncontour_2(fig, minuit):
-    minuit.draw_mncontour("x", "y", cl=0.68)
+    # use 0, 1 instead of "x", "y"
+    minuit.draw_mncontour(0, 1, cl=0.68)
 
 
 def test_mncontour_3(fig, minuit):
@@ -95,7 +81,8 @@ def test_contour_1(fig, minuit):
 
 
 def test_contour_2(fig, minuit):
-    minuit.draw_contour("x", "y", size=20, bound=2)
+    # use 0, 1 instead of "x", "y"
+    minuit.draw_contour(0, 1, size=20, bound=2)
 
 
 def test_contour_3(fig, minuit):
@@ -148,7 +135,10 @@ def test_mnmatrix_7(fig):
     m.draw_mnmatrix(cl=[1, 3])
 
 
+@pytest.mark.filterwarnings("ignore::DeprecationWarning")
 def test_interactive():
+    ipywidgets = pytest.importorskip("ipywidgets")
+
     def cost(a, b):
         return a**2 + b**2
 
@@ -170,81 +160,68 @@ def test_interactive():
 
     plot = Plot()
 
-    try:
-        import ipywidgets  # noqa
-        import IPython  # noqa
+    m = Minuit(cost, 1, 1)
+    with pytest.raises(AttributeError, match="no visualize method"):
+        m.interactive(raise_on_exception=True)
 
-        ipywidgets_available = True
+    with plot.assert_call():
+        out1 = m.interactive(plot)
+    assert isinstance(out1, ipywidgets.HBox)
 
-        m = Minuit(cost, 1, 1)
-        with pytest.raises(ValueError, match="no visualize method"):
-            m.interactive(raise_on_exception=True)
+    # manipulate state to also check this code
+    ui = out1.children[1]
+    header, parameters = ui.children
+    fit_button, update_button, reset_button, algo_select = header.children
+    with plot.assert_call():
+        fit_button.click()
+    assert_allclose(m.values, (0, 0), atol=1e-5)
+    with plot.assert_call():
+        reset_button.click()
+    assert_allclose(m.values, (1, 1), atol=1e-5)
 
-        with plot.assert_call():
-            out1 = m.interactive(plot)
-        assert isinstance(out1, ipywidgets.HBox)
+    algo_select.value = "Scipy"
+    with plot.assert_call():
+        fit_button.click()
 
-        # manipulate state to also check this code
-        ui = out1.children[1]
-        header, parameters = ui.children
-        fit_button, update_button, reset_button, algo_select = header.children
-        with plot.assert_call():
-            fit_button.click()
-        assert_allclose(m.values, (0, 0), atol=1e-5)
-        with plot.assert_call():
-            reset_button.click()
-        assert_allclose(m.values, (1, 1), atol=1e-5)
+    algo_select.value = "Simplex"
+    with plot.assert_call():
+        fit_button.click()
 
-        algo_select.value = "Scipy"
-        with plot.assert_call():
-            fit_button.click()
+    update_button.value = False
+    with plot.assert_call():
+        parameters.children[0].slider.value = 0.4  # change first slider
+    parameters.children[0].fix.value = True
+    with plot.assert_call():
+        parameters.children[0].opt.value = True
 
-        algo_select.value = "Simplex"
-        with plot.assert_call():
-            fit_button.click()
+    class Cost:
+        def visualize(self, args):
+            return plot(args)
 
-        update_button.value = False
-        with plot.assert_call():
-            parameters.children[0].slider.value = 0.4  # change first slider
-        parameters.children[0].fix.value = True
-        with plot.assert_call():
-            parameters.children[0].opt.value = True
+        def __call__(self, a, b):
+            return (a - 100) ** 2 + (b + 100) ** 2
 
-        class Cost:
-            def visualize(self, args):
-                return plot(args)
+    c = Cost()
+    m = Minuit(c, 0, 0)
+    with plot.assert_call():
+        out = m.interactive(raise_on_exception=True)
 
-            def __call__(self, a, b):
-                return (a - 100) ** 2 + (b + 100) ** 2
-
-        c = Cost()
-        m = Minuit(c, 0, 0)
-        with plot.assert_call():
-            out = m.interactive(raise_on_exception=True)
-
-        # this should modify slider range
-        ui = out.children[1]
-        header, parameters = ui.children
-        fit_button, update_button, reset_button, algo_select = header.children
-        assert parameters.children[0].slider.max < 100
-        assert parameters.children[1].slider.min > -100
-        with plot.assert_call():
-            fit_button.click()
-        assert_allclose(m.values, (100, -100), atol=1e-5)
-        # this should trigger an exception
-        plot.raises = True
-        with plot.assert_call():
-            fit_button.click()
-
-    except ModuleNotFoundError:
-        ipywidgets_available = False
-
-    if not ipywidgets_available:
-        with pytest.raises(ModuleNotFoundError, match="Please install"):
-            m = Minuit(cost, 1, 1)
-            m.interactive()
+    # this should modify slider range
+    ui = out.children[1]
+    header, parameters = ui.children
+    fit_button, update_button, reset_button, algo_select = header.children
+    assert parameters.children[0].slider.max < 100
+    assert parameters.children[1].slider.min > -100
+    with plot.assert_call():
+        fit_button.click()
+    assert_allclose(m.values, (100, -100), atol=1e-5)
+    # this should trigger an exception
+    plot.raises = True
+    with plot.assert_call():
+        fit_button.click()
 
 
+@pytest.mark.filterwarnings("ignore::DeprecationWarning")
 def test_interactive_raises():
     pytest.importorskip("ipywidgets")
 
@@ -260,6 +237,7 @@ def test_interactive_raises():
         m.interactive(raiser, raise_on_exception=True)
 
 
+@pytest.mark.filterwarnings("ignore::DeprecationWarning")
 def test_interactive_with_array_func():
     pytest.importorskip("ipywidgets")
 
