@@ -3,7 +3,9 @@ import pytest
 import numpy as np
 from numpy.testing import assert_allclose, assert_equal
 from iminuit import Minuit
-from iminuit.util import Param, IMinuitWarning, make_func_code
+from iminuit.util import Param, make_func_code
+from iminuit.warnings import IMinuitWarning
+from iminuit.typing import Annotated
 from pytest import approx
 from argparse import Namespace
 
@@ -23,24 +25,12 @@ def debug():
 is_pypy = platform.python_implementation() == "PyPy"
 
 
-def test_pedantic_warning_message():
-    with pytest.warns(IMinuitWarning, match=r"errordef not set, using 1"):
-        m = Minuit(lambda x: 0, x=0)
-        m.migrad()  # MARKER
-
-
 def test_version():
     import iminuit
 
     assert iminuit.__version__
 
 
-def lsq(func):
-    func.errordef = Minuit.LEAST_SQUARES
-    return func
-
-
-@lsq
 def func0(x, y):  # values = (2.0, 5.0), errors = (2.0, 1.0)
     return (x - 2.0) ** 2 / 4.0 + np.exp((y - 5.0) ** 2) + 10
 
@@ -68,7 +58,6 @@ class Func2:
         return func0(arg[0], arg[1]) * 4
 
 
-@lsq
 def func4(x, y, z):
     return 0.2 * (x - 2.0) ** 2 + 0.1 * (y - 5.0) ** 2 + 0.25 * (z - 7.0) ** 2 + 10
 
@@ -80,7 +69,6 @@ def func4_grad(x, y, z):
     return dfdx, dfdy, dfdz
 
 
-@lsq
 def func5(x, long_variable_name_really_long_why_does_it_has_to_be_this_long, z):
     return (
         (x - 1) ** 2
@@ -96,14 +84,11 @@ def func5_grad(x, long_variable_name_really_long_why_does_it_has_to_be_this_long
     return dfdx, dfdy, dfdz
 
 
-@lsq
 def func6(x, m, s, a):
     return a / ((x - m) ** 2 + s**2)
 
 
 class Correlated:
-    errordef = 1
-
     def __init__(self):
         sx = 2
         sy = 1
@@ -115,7 +100,6 @@ class Correlated:
         return np.dot(x.T, np.dot(self.cinv, x))
 
 
-@lsq
 def func_np(x):  # test numpy support
     return np.sum((x - 1) ** 2)
 
@@ -172,15 +156,36 @@ def func_test_helper(f, grad=None, errordef=None):
     return m
 
 
-def test_func0():  # check that providing gradient improves convergence
+def test_mncontour_interpolated_1():
+    m = Minuit(func0, 1, 1)
+    m.migrad()
+
+    # interpolated < size is ignored
+    pts = m.mncontour("x", "y", size=20, interpolated=10)
+    assert len(pts) == 21
+
+
+def test_mncontour_interpolated_2():
+    pytest.importorskip("scipy.interpolate")
+
+    m = Minuit(func0, 1, 1)
+    m.migrad()
+
+    pts = m.mncontour("x", "y", size=20, interpolated=200)
+    assert len(pts) == 200
+
+
+def test_func0():
     m1 = func_test_helper(func0)
     m2 = func_test_helper(func0, grad=func0_grad)
     assert m1.ngrad == 0
     assert m2.ngrad > 0
+    # check that providing gradient improves convergence
+    assert m2.nfcn < m1.nfcn
 
 
 def test_lambda():
-    func_test_helper(lambda x, y: func0(x, y), errordef=1)
+    func_test_helper(lambda x, y: func0(x, y))
 
 
 def test_Func1():
@@ -188,7 +193,8 @@ def test_Func1():
 
 
 def test_Func2():
-    func_test_helper(Func2())
+    with pytest.warns(np.VisibleDeprecationWarning):
+        func_test_helper(Func2())
 
 
 def test_no_signature():
@@ -196,8 +202,6 @@ def test_no_signature():
         x, y = args
         return (x - 1) ** 2 + (y - 2) ** 2
 
-    no_signature.errordef = 1
-
     m = Minuit(no_signature, 3, 4)
     assert m.values == (3, 4)
     assert m.parameters == ("x0", "x1")
@@ -293,6 +297,7 @@ def test_array_func_2():
     assert m.fixed == (False, True)
     assert m.limits["a"] == (0, 2)
     m.migrad()
+    assert m.fmin.ngrad > 0
     assert_allclose(m.values, (1, 1), rtol=1e-2)
     c = m.covariance
     assert_allclose(c, ((1, 0), (0, 0)), rtol=1e-2)
@@ -304,7 +309,6 @@ def test_array_func_2():
 
 def test_wrong_use_of_array_init():
     m = Minuit(lambda a, b: a**2 + b**2, (1, 2))
-    m.errordef = Minuit.LEAST_SQUARES
     with pytest.raises(TypeError):
         m.migrad()
 
@@ -353,7 +357,7 @@ def test_initial_guesses():
 
 
 @pytest.mark.parametrize("grad", (None, func0_grad))
-def test_fix_param(grad):
+def test_fixed(grad):
     m = Minuit(func0, grad=grad, x=0, y=0)
     assert m.npar == 2
     assert m.nfit == 2
@@ -363,9 +367,10 @@ def test_fix_param(grad):
     assert_allclose(m.errors, (2, 1), rtol=1e-4)
     assert_allclose(m.covariance, ((4, 0), (0, 1)), atol=1e-4)
 
-    # now fix y = 10
-    m = Minuit(func0, grad=grad, x=0, y=10.0)
+    m = Minuit(func0, grad=grad, x=0, y=10)
+    assert not m.fixed["y"]
     m.fixed["y"] = True
+    assert m.fixed["y"]
     assert m.npar == 2
     assert m.nfit == 1
     m.migrad()
@@ -410,6 +415,29 @@ def test_fix_param(grad):
     assert m.fixed[:1] == [True]
 
 
+def test_fixto():
+    m = Minuit(func0, x=0, y=0)
+    assert np.all(~m.fixed)
+    m.fixto(0, 1)
+    assert m.fixed[0]
+    assert m.values[0] == 1
+    m.fixto([0, 1], 0)
+    assert np.all(m.fixed)
+    assert m.values == [0, 0]
+    m.fixed = False
+    assert np.all(~m.fixed)
+    m.fixto(slice(0, 2), [1, 2])
+    assert np.all(m.fixed)
+    assert_equal(m.values, [1, 2])
+    m.fixed = False
+    assert np.all(~m.fixed)
+    m.fixto(..., [2, 3])
+    assert np.all(m.fixed)
+    assert_equal(m.values, [2, 3])
+    with pytest.raises(ValueError, match="length of argument"):
+        m.fixto([1], [1, 2])
+
+
 @pytest.mark.parametrize("grad", (None, func0_grad))
 def test_minos(grad):
     m = Minuit(func0, grad=grad, x=0, y=0)
@@ -422,7 +450,7 @@ def test_minos(grad):
     assert m.merrors[-1].upper == m.merrors["y"].upper
 
 
-@pytest.mark.parametrize("cl", (0.68, 0.90))
+@pytest.mark.parametrize("cl", (0.68, 0.90, 1, 1.5, 2))
 @pytest.mark.parametrize("k", (10, 1000))
 @pytest.mark.parametrize("limit", (False, True))
 def test_minos_cl(cl, k, limit):
@@ -434,10 +462,15 @@ def test_minos_cl(cl, k, limit):
 
     # find location of min + up by hand
     def crossing(x):
-        up = 0.5 * stats.chi2(1).ppf(cl)
         return nll(k + x) - (nll(k) + up)
 
-    bound = 1.5 * (stats.chi2(1).ppf(cl) * k) ** 0.5
+    if cl >= 1:
+        bound = cl * k**0.5
+        up = 0.5 * cl**2
+    else:
+        bound = (stats.chi2(1).ppf(cl) * k) ** 0.5
+        up = 0.5 * stats.chi2(1).ppf(cl)
+    bound *= 1.5
     upper = opt.root_scalar(crossing, bracket=(0, bound)).root
     lower = opt.root_scalar(crossing, bracket=(-bound, 0)).root
 
@@ -495,7 +528,7 @@ def test_minos_single_fixed():
     m = Minuit(func0, x=0, y=0)
     m.fixed["x"] = True
     m.migrad()
-    m.minos("y")
+    m.minos(1)
     assert len(m.merrors) == 1
     me = m.merrors["y"]
     assert me.name == "y"
@@ -524,18 +557,24 @@ def test_minos_single_no_migrad():
 def test_minos_single_nonsense_variable():
     m = Minuit(func0, x=0, y=0)
     m.migrad()
-    with pytest.raises(RuntimeError):
+    with pytest.raises(ValueError):
         m.minos("nonsense")
 
 
 def test_minos_with_bad_fmin():
     m = Minuit(lambda x: 0, x=0)
-    m.errordef = 1
     m.migrad()
     with pytest.raises(RuntimeError):
         m.minos()
 
 
+def test_minos_bad_index():
+    m = Minuit(func0, 1, 1)
+    m.migrad()
+    with pytest.raises(ValueError):
+        m.minos(2)
+
+
 @pytest.mark.parametrize("grad", (None, func5_grad))
 def test_fixing_long_variable_name(grad):
     m = Minuit(
@@ -545,7 +584,6 @@ def test_fixing_long_variable_name(grad):
         x=0,
         z=0,
     )
-    m.errordef = 1
     m.fixed["long_variable_name_really_long_why_does_it_has_to_be_this_long"] = True
     m.migrad()
     assert_allclose(m.values, [1, 2, -1], atol=1e-3)
@@ -578,16 +616,22 @@ def test_initial_value():
 
 
 @pytest.mark.parametrize("grad", (None, func0_grad))
-@pytest.mark.parametrize("cl", (None, 0.5, 0.9))
-def test_mncontour(grad, cl):
+@pytest.mark.parametrize("cl", (None, 0.5, 0.9, 1, 1.5, 2))
+@pytest.mark.parametrize("experimental", (False, True))
+def test_mncontour(grad, cl, experimental):
     stats = pytest.importorskip("scipy.stats")
+
     m = Minuit(func0, grad=grad, x=1.0, y=2.0)
     m.migrad()
-    ctr = m.mncontour("x", "y", size=30, cl=cl)
+    ctr = m.mncontour("x", "y", size=30, cl=cl, experimental=experimental)
 
-    factor = stats.chi2(2).ppf(0.68 if cl is None else cl)
+    if cl is None:
+        cl = 0.68
+    elif cl >= 1:
+        cl = stats.chi2(1).cdf(cl**2)
+    factor = stats.chi2(2).ppf(cl)
     cl2 = stats.chi2(1).cdf(factor)
-    assert len(ctr) == 30
+    assert len(ctr) == 31
     assert len(ctr[0]) == 2
 
     m.minos("x", "y", cl=cl2)
@@ -598,8 +642,74 @@ def test_mncontour(grad, cl):
     cmax = np.max(ctr, axis=0)
 
     x, y = m.values
-    assert_allclose((x + xm.lower, y + ym.lower), cmin)
-    assert_allclose((x + xm.upper, y + ym.upper), cmax)
+    assert_allclose((x + xm.lower, y + ym.lower), cmin, atol=1e-2)
+    assert_allclose((x + xm.upper, y + ym.upper), cmax, atol=1e-2)
+
+
+@pytest.mark.parametrize("experimental", (False, True))
+def test_mncontour_limits(experimental):
+    pytest.importorskip("scipy.optimize")
+
+    def cost(x, y):
+        return x**2 + y**2
+
+    m = Minuit(cost, x=0.5, y=0.5)
+    m.limits = (0, 2)
+    m.migrad()
+    cont = m.mncontour(0, 1, size=30, experimental=experimental)
+
+    assert np.all(cont[:, 0] >= 0)
+    assert np.all(cont[:, 1] >= 0)
+
+
+def test_mncontour_no_fmin():
+    m = Minuit(func0, x=0, y=0)
+
+    with pytest.raises(RuntimeError):
+        # fails, because this is not a minimum
+        m.mncontour("x", "y")
+
+    # succeeds
+    m.values = (2, 5)
+    # use 0, 1 instead of "x", "y"
+    c = m.mncontour(0, 1, size=10)
+
+    # compute reference to compare with
+    m2 = Minuit(func0, x=0, y=0)
+    m2.migrad()
+    c2 = m.mncontour("x", "y", size=10)
+
+    assert_allclose(c, c2)
+
+
+def test_mncontour_with_fixed_var():
+    m = Minuit(func0, x=0, y=0)
+    m.fixed["x"] = True
+    m.migrad()
+    with pytest.raises(ValueError):
+        m.mncontour("x", "y")
+
+
+@pytest.mark.parametrize("experimental", (False, True))
+def test_mncontour_array_func(experimental):
+    stats = pytest.importorskip("scipy.stats")
+
+    m = Minuit(Correlated(), (0, 0), name=("x", "y"))
+    m.migrad()
+
+    cl = stats.chi2(2).cdf(1)
+    ctr = m.mncontour("x", "y", size=30, cl=cl, experimental=experimental)
+    assert len(ctr) == 31
+    assert len(ctr[0]) == 2
+
+    m.minos("x", "y")
+    x, y = m.values
+    xm = m.merrors["x"]
+    ym = m.merrors["y"]
+    cmin = np.min(ctr, axis=0)
+    cmax = np.max(ctr, axis=0)
+    assert_allclose((x + xm.lower, y + ym.lower), cmin, atol=1e-2)
+    assert_allclose((x + xm.upper, y + ym.upper), cmax, atol=1e-2)
 
 
 @pytest.mark.parametrize("grad", (None, func0_grad))
@@ -611,6 +721,42 @@ def test_contour(grad):
     assert_allclose(func0(X, Y), v.T)
 
 
+def test_contour_separate_size():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    x, y, v = m.contour("x", "y", size=(10, 20))
+    assert len(x) == 10
+    assert len(y) == 20
+    X, Y = np.meshgrid(x, y)
+    assert_allclose(func0(X, Y), v.T)
+
+
+def test_contour_grid():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    x, y, v = m.contour("x", "y", grid=(np.linspace(0, 2, 10), np.linspace(0, 4, 20)))
+    assert len(x) == 10
+    assert len(y) == 20
+    X, Y = np.meshgrid(x, y)
+    assert_allclose(func0(X, Y), v.T)
+
+
+def test_contour_bad_grid():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    with pytest.raises(ValueError):
+        m.contour("x", "y", grid=([1, 2, 3], [[1, 2, 3]]))
+
+    with pytest.raises(ValueError):
+        m.contour("x", "y", grid=([1, 2, 3],))
+
+    with pytest.raises(ValueError):
+        m.contour("x", "y", grid=([1, 2, 3], [1, 2], [3, 4]))
+
+    with pytest.raises(ValueError):
+        m.contour("x", "y", grid=(10, [1, 2, 3]))
+
+
 @pytest.mark.parametrize("grad", (None, func0_grad))
 def test_profile(grad):
     m = Minuit(func0, grad=grad, x=1.0, y=2.0)
@@ -624,6 +770,26 @@ def test_profile(grad):
     assert_allclose(v - np.min(v), v2)
 
 
+def test_profile_grid():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    y, v = m.profile("y", grid=np.linspace(0, 4, 15))
+    assert len(y) == 15
+    assert y[0] == 0
+    assert y[-1] == 4
+    assert_allclose(func0(m.values[0], y), v)
+
+
+def test_profile_bad_grid():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    with pytest.raises(ValueError):
+        m.profile("y", grid=[[1, 2, 3]])
+
+    with pytest.raises(ValueError):
+        m.profile("y", grid=10)
+
+
 @pytest.mark.parametrize("grad", (None, func0_grad))
 def test_mnprofile(grad):
     m = Minuit(func0, grad=grad, x=1.0, y=2.0)
@@ -643,83 +809,66 @@ def test_mnprofile(grad):
 
     assert_allclose(v, v2)
 
-    y, v3, _ = m.mnprofile("y", size=10, subtract_min=True)
+    # use 1 instead of "y"
+    y, v3, _ = m.mnprofile(1, size=10, subtract_min=True)
     assert np.min(v3) == 0
     assert_allclose(v - np.min(v), v3)
 
 
+def test_mnprofile_grid():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    y, v, _ = m.mnprofile("y", grid=np.linspace(0, 4, 15))
+    assert len(y) == 15
+    assert y[0] == 0
+    assert y[-1] == 4
+    m2 = Minuit(func0, x=1.0, y=2.0)
+    m2.fixed[1] = True
+    v2 = []
+    for yi in y:
+        m2.values = (m.values[0], yi)
+        m2.migrad()
+        v2.append(m2.fval)
+    assert_allclose(v, v2)
+
+
+def test_mnprofile_bad_grid():
+    m = Minuit(func0, x=1.0, y=2.0)
+    m.migrad()
+    with pytest.raises(ValueError):
+        m.mnprofile("y", grid=10)
+    with pytest.raises(ValueError):
+        m.mnprofile("y", grid=[[10, 20]])
+
+
 def test_contour_subtract():
     m = Minuit(func0, x=1.0, y=2.0)
     m.migrad()
     v = m.contour("x", "y", subtract_min=False)[2]
-    v2 = m.contour("x", "y", subtract_min=True)[2]
+    # use 0 and 1 instead "x", "y"
+    v2 = m.contour(0, 1, subtract_min=True)[2]
     assert np.min(v2) == 0
     assert_allclose(v - np.min(v), v2)
 
 
-def test_mncontour_no_fmin():
-    m = Minuit(func0, x=0, y=0)
-
-    with pytest.raises(RuntimeError):
-        m.mncontour("x", "y")  # fails, because this is not a minimum
-
-    # succeeds
-    m.values = (2, 5)
-    c = m.mncontour("x", "y", size=10)
-
-    # compute reference to compare with
-    m2 = Minuit(func0, x=0, y=0)
-    m2.migrad()
-    c2 = m.mncontour("x", "y", size=10)
-
-    assert_allclose(c, c2)
-
-
-def test_mncontour_with_fixed_var():
-    m = Minuit(func0, x=0, y=0)
-    m.errordef = 1
-    m.fixed["x"] = True
-    m.migrad()
-    with pytest.raises(ValueError):
-        m.mncontour("x", "y")
-
-
-def test_mncontour_array_func():
-    stats = pytest.importorskip("scipy.stats")
-
-    m = Minuit(Correlated(), (0, 0), name=("x", "y"))
-    m.migrad()
-
-    cl = stats.chi2(2).cdf(1)
-    ctr = m.mncontour("x", "y", size=30, cl=cl)
-    assert len(ctr) == 30
-    assert len(ctr[0]) == 2
-
-    m.minos("x", "y")
-    x, y = m.values
-    xm = m.merrors["x"]
-    ym = m.merrors["y"]
-    cmin = np.min(ctr, axis=0)
-    cmax = np.max(ctr, axis=0)
-    assert_allclose((x + xm.lower, y + ym.lower), cmin)
-    assert_allclose((x + xm.upper, y + ym.upper), cmax)
-
-
 def test_profile_array_func():
     m = Minuit(Correlated(), (0, 0), name=("x", "y"))
     m.migrad()
-    m.profile("y")
+    a = m.profile("y")
+    b = m.profile(1)
+    assert_equal(a, b)
 
 
 def test_mnprofile_array_func():
     m = Minuit(Correlated(), (0, 0), name=("x", "y"))
     m.migrad()
-    m.mnprofile("y")
+    a = m.mnprofile("y")
+    b = m.mnprofile(1)
+    assert_equal(a, b)
 
 
 def test_mnprofile_bad_func():
     m = Minuit(lambda x, y: 0, 0, 0)
-    m.errordef = 1
     with pytest.warns(IMinuitWarning):
         m.mnprofile("x")
 
@@ -793,7 +942,6 @@ def test_values(minuit):
 def test_fmin():
     m = Minuit(lambda x, s: (x * s) ** 2, x=1, s=1)
     m.fixed["s"] = True
-    m.errordef = 1
     m.migrad()
     fm1 = m.fmin
     assert fm1.is_valid
@@ -812,7 +960,6 @@ def test_chi2_fit():
         return (x - 1) ** 2 + ((y - 2) / 3) ** 2
 
     m = Minuit(chi2, x=0, y=0)
-    m.errordef = 1
     m.migrad()
     assert_allclose(m.values, (1, 2))
     assert_allclose(m.errors, (1, 3))
@@ -889,7 +1036,6 @@ def test_oneside_outside():
 def test_migrad_ncall():
     class Func:
         nfcn = 0
-        errordef = 1
 
         def __call__(self, x):
             self.nfcn += 1
@@ -916,21 +1062,23 @@ def test_migrad_ncall():
     assert m.nfcn < ncalls_without_limit
 
 
-def test_ngrad():
+@pytest.mark.parametrize("arg", (1, np.array([1.0, 2.0])))
+def test_ngrad(arg):
     class Func:
-        errordef = 1
         ngrad = 0
 
         def __call__(self, x):
-            return x**2
+            return np.sum(x**2)
 
         def grad(self, x):
             self.ngrad += 1
+            if np.ndim(x) == 1:
+                return 2 * x
             return [2 * x]
 
     # check that counting is accurate
     fcn = Func()
-    m = Minuit(fcn, 1)
+    m = Minuit(fcn, arg)
     m.migrad()
     assert m.ngrad > 0
     assert m.ngrad == fcn.ngrad
@@ -944,6 +1092,15 @@ def test_ngrad():
     m.hesse()
     assert m.ngrad == before
 
+    m.reset()
+    m.migrad()
+    m2 = Minuit(lambda x: fcn(x), arg)
+    m2.migrad()
+    assert m.ngrad > 0
+    assert m2.ngrad == 0
+    # apparently this is not always the case:
+    # assert m2.nfcn > m.nfcn
+
 
 def test_errordef():
     m = Minuit(lambda x: x**2, 0)
@@ -973,7 +1130,6 @@ def test_print_level():
 
 def test_params():
     m = Minuit(func0, x=1, y=2)
-    m.errordef = Minuit.LEAST_SQUARES
     m.errors = (3, 4)
     m.fixed["x"] = True
     m.limits["y"] = (None, 10)
@@ -1006,7 +1162,6 @@ def test_params():
 
 def test_non_analytical_function():
     class Func:
-        errordef = 1
         i = 0
 
         def __call__(self, a):
@@ -1015,13 +1170,12 @@ def test_non_analytical_function():
 
     m = Minuit(Func(), 0)
     m.migrad()
-    assert m.fmin.is_valid is False
-    assert m.fmin.is_above_max_edm is True
+    assert not m.fmin.is_valid
+    assert m.fmin.is_above_max_edm
 
 
 def test_non_invertible():
     m = Minuit(lambda x, y: 0, 1, 2)
-    m.errordef = 1
     m.strategy = 0
     m.migrad()
     assert m.fmin.is_valid
@@ -1032,10 +1186,9 @@ def test_non_invertible():
 
 def test_function_without_local_minimum():
     m = Minuit(lambda a: -a, 0)
-    m.errordef = 1
     m.migrad()
-    assert m.fmin.is_valid is False
-    assert m.fmin.is_above_max_edm is True
+    assert not m.fmin.is_valid
+    assert m.fmin.is_above_max_edm
 
 
 def test_function_with_maximum():
@@ -1043,9 +1196,8 @@ def test_function_with_maximum():
         return -(a**2)
 
     m = Minuit(func, a=0)
-    m.errordef = 1
     m.migrad()
-    assert m.fmin.is_valid is False
+    assert not m.fmin.is_valid
 
 
 def test_perfect_correlation():
@@ -1053,12 +1205,11 @@ def test_perfect_correlation():
         return (a - b) ** 2
 
     m = Minuit(func, a=1, b=2)
-    m.errordef = 1
     m.migrad()
-    assert m.fmin.is_valid is True
-    assert m.fmin.has_accurate_covar is False
-    assert m.fmin.has_posdef_covar is False
-    assert m.fmin.has_made_posdef_covar is True
+    assert m.fmin.is_valid
+    assert not m.fmin.has_accurate_covar
+    assert not m.fmin.has_posdef_covar
+    assert m.fmin.has_made_posdef_covar
 
 
 def test_modify_param_state():
@@ -1102,7 +1253,6 @@ def test_hesse_without_migrad():
     assert m.fmin
 
     m = Minuit(lambda x: 0, 0)
-    m.errordef = 1
     m.hesse()
     assert not m.accurate
     assert m.fmin.hesse_failed
@@ -1143,7 +1293,6 @@ def returning_garbage(x):
 )
 def test_bad_functions(func, expected):
     m = Minuit(func, x=1)
-    m.errordef = 1
     m.throw_nan = True
     with pytest.raises(type(expected)) as excinfo:
         m.migrad()
@@ -1152,7 +1301,6 @@ def test_bad_functions(func, expected):
 
 def test_throw_nan():
     m = Minuit(returning_nan, x=1)
-    m.errordef = 1
     assert not m.throw_nan
     m.migrad()
     m.throw_nan = True
@@ -1185,7 +1333,6 @@ def returning_noniterable(x):
 )
 def test_bad_functions_np(func, expected):
     m = Minuit(lambda x: np.dot(x, x), (1, 1), grad=func)
-    m.errordef = 1
     m.throw_nan = True
     with pytest.raises(type(expected)) as excinfo:
         m.migrad()
@@ -1195,17 +1342,15 @@ def test_bad_functions_np(func, expected):
 @pytest.mark.parametrize("sign", (-1, 1))
 def test_parameter_at_limit(sign):
     m = Minuit(lambda x: (x - sign * 1.2) ** 2, x=0)
-    m.errordef = 1
     m.limits["x"] = (-1, 1)
     m.migrad()
     assert m.values["x"] == approx(sign * 1.0, abs=1e-3)
-    assert m.fmin.has_parameters_at_limit is True
+    assert m.fmin.has_parameters_at_limit
 
     m = Minuit(lambda x: (x - sign * 1.2) ** 2, x=0)
-    m.errordef = 1
     m.migrad()
     assert m.values["x"] == approx(sign * 1.2, abs=1e-3)
-    assert m.fmin.has_parameters_at_limit is False
+    assert not m.fmin.has_parameters_at_limit
 
 
 @pytest.mark.parametrize("iterate,valid", ((1, False), (5, True)))
@@ -1214,20 +1359,17 @@ def test_inaccurate_fcn(iterate, valid):
         return abs(x) ** 10 + 1e7
 
     m = Minuit(f, x=2)
-    m.errordef = 1
     m.migrad(iterate=iterate)
     assert m.valid == valid
 
 
 def test_migrad_iterate():
     m = Minuit(lambda x: 0, x=2)
-    m.errordef = 1
     with pytest.raises(ValueError):
         m.migrad(iterate=0)
 
 
 def test_precision():
-    @lsq
     def fcn(x):
         return np.exp(x * x + 1)
 
@@ -1349,7 +1491,6 @@ def test_cfunc():
 
     c_sig = nb.types.double(nb.types.uintc, nb.types.CPointer(nb.types.double))
 
-    @lsq
     @nb.cfunc(c_sig)
     def fcn(n, x):
         x = nb.carray(x, (n,))
@@ -1364,7 +1505,8 @@ def test_cfunc():
 
 
 @pytest.mark.parametrize("cl", (0.5, None, 0.9))
-def test_confidence_level(cl):
+@pytest.mark.parametrize("experimental", (False, True))
+def test_confidence_level(cl, experimental):
     stats = pytest.importorskip("scipy.stats")
     mpath = pytest.importorskip("matplotlib.path")
 
@@ -1386,7 +1528,7 @@ def test_confidence_level(cl):
     r = d.rvs(n, random_state=1)
 
     # check that mncontour indeed contains fraction of random points equal to CL
-    pts = m.mncontour("x0", "x1", cl=cl)
+    pts = m.mncontour("x0", "x1", cl=cl, experimental=experimental)
     p = mpath.Path(pts)
     cl2 = np.sum(p.contains_points(r)) / n
     assert cl2 == approx(cl_ref, abs=0.01)
@@ -1453,7 +1595,7 @@ def test_pickle(grad):
 
 def test_minos_new_min():
     xref = [1.0]
-    m = Minuit(lsq(lambda x: (x - xref[0]) ** 2), x=0)
+    m = Minuit(lambda x: (x - xref[0]) ** 2, x=0)
     m.migrad()
     assert m.values[0] == approx(xref[0], abs=1e-3)
     m.minos()
@@ -1469,7 +1611,7 @@ def test_minos_new_min():
 
 
 def test_minos_without_migrad():
-    m = Minuit(lsq(lambda x, y: (x - 1) ** 2 + (y / 2) ** 2), 1.001, 0.001)
+    m = Minuit(lambda x, y: (x - 1) ** 2 + (y / 2) ** 2, 1.001, 0.001)
     m.minos()
     me = m.merrors["x"]
     assert me.is_valid
@@ -1482,18 +1624,72 @@ def test_minos_without_migrad():
 
 
 def test_missing_ndata():
-    def fcn(a):
-        return a
-
-    fcn.errordef = 1
-
-    m = Minuit(fcn, 1)
+    m = Minuit(lambda a: a, 1)
     assert_equal(m.ndof, np.nan)
 
 
 def test_call_limit_reached_in_hesse():
     m = Minuit(lambda x: ((x - 1.2) ** 4).sum(), np.ones(10) * 10)
-    m.errordef = 1
     m.migrad(ncall=200)
     assert m.fmin.has_reached_call_limit
     assert m.fmin.nfcn < 205
+
+
+def test_bad_cl():
+    m = Minuit(func0, 1, 1)
+    m.migrad()
+
+    for cl in (0, -1):
+        with pytest.raises(ValueError):
+            m.minos(cl=cl)
+
+        with pytest.raises(ValueError):
+            m.mncontour("x", "y", cl=cl)
+
+
+def test_negative_errors():
+    m = Minuit(func0, -1, -1)
+    assert np.all(np.array(m.errors) > 0)
+    with pytest.warns():
+        m.errors[0] = -1
+    assert m.errors[0] > 0
+    with pytest.warns():
+        m.errors = -2
+    assert np.all(np.array(m.errors) > 0)
+    m.errors = 10
+    assert_allclose(m.errors, 10)
+    m.errors = (1, 2)
+    assert_allclose(m.errors, (1, 2))
+
+
+def test_visualize():
+    m = Minuit(func0, 1, 1)
+    m.migrad()
+    with pytest.raises(AttributeError):
+        m.visualize()
+
+    kwargs = {}
+
+    func0.visualize = lambda args, **kw: kwargs.update(kw)
+    m.visualize(foo="bar")
+
+    assert kwargs == {"foo": "bar"}
+
+    del func0.visualize
+
+
+def test_annotated_cost_function():
+    def cost(a, b: Annotated[float, 0.1:1]):
+        return a**2 + b**2
+
+    m = Minuit(cost, 0.5, 0.5)
+    assert m.limits[0] == (-np.inf, np.inf)
+    assert m.limits[1] == (0.1, 1.0)
+    m.migrad()
+    assert_allclose(m.values, (0, 0.1), atol=1e-2)
+
+    m2 = Minuit(cost, 0.5, 0.5, name=("x", "y"))
+    assert m2.limits["x"] == (-np.inf, np.inf)
+    assert m2.limits["y"] == (0.1, 1.0)
+    m.migrad()
+    assert_allclose(m.values, (0, 0.1), atol=1e-2)
