@@ -4,7 +4,16 @@ from argparse import Namespace
 from numpy.testing import assert_equal, assert_allclose
 import numpy as np
 from iminuit._core import MnUserParameterState
+from iminuit._optional_dependencies import optional_module_for
 import pickle
+from iminuit._hide_modules import hide_modules
+
+try:
+    import scipy  # noqa
+
+    scipy_available = True
+except ModuleNotFoundError:
+    scipy_available = False
 
 
 def test_ndim():
@@ -122,10 +131,10 @@ def test_FixedView_comparison_with_broadcasting():
     assert_equal(f, [False, False, False])
 
     # broadcasting
-    assert f == False
+    assert f == False  # noqa
     f[0] = True
     assert_equal(f, [True, False, False])
-    assert f != False
+    assert f != False  # noqa
 
 
 def test_Matrix():
@@ -205,11 +214,11 @@ def test_Param():
     assert p.value == 1.2
     assert p.error == 3.4
     assert p.merror is None
-    assert p.is_const == False
-    assert p.is_fixed == False
-    assert p.has_limits == True
-    assert p.has_lower_limit is True
-    assert p.has_upper_limit is False
+    assert not p.is_const
+    assert not p.is_fixed
+    assert p.has_limits
+    assert p.has_lower_limit
+    assert not p.has_upper_limit
     assert p.lower_limit == 42
     assert p.upper_limit is None
 
@@ -364,7 +373,7 @@ def test_FMin(errordef):
     assert fmin.algorithm == "foo"
     assert fmin.edm == 1.23456e-10
     assert fmin.edm_goal == 0.1
-    assert fmin.has_parameters_at_limit == False
+    assert not fmin.has_parameters_at_limit
     assert fmin.time == 1.2
 
     assert fmin == util.FMin(fm, "foo", 1, 2, 1, 0.1, 1.2)
@@ -441,11 +450,13 @@ def test_address_of_cfunc_bad_signature():
 
 
 def test_make_func_code():
-    fc = util.make_func_code(["a", "b"])
+    with pytest.warns(np.VisibleDeprecationWarning):
+        fc = util.make_func_code(["a", "b"])
     assert fc.co_varnames == ("a", "b")
     assert fc.co_argcount == 2
 
-    fc = util.make_func_code(("x",))
+    with pytest.warns(np.VisibleDeprecationWarning):
+        fc = util.make_func_code(("x",))
     assert fc.co_varnames == ("x",)
     assert fc.co_argcount == 1
 
@@ -507,6 +518,7 @@ def test_merge_signatures():
     assert pg == (0, 3, 4)
 
 
+@pytest.mark.skipif(not scipy_available, reason="needs scipy")
 def test_propagate_1():
     cov = [
         [1.0, 0.1, 0.2],
@@ -521,14 +533,17 @@ def test_propagate_1():
     with pytest.warns(np.VisibleDeprecationWarning):
         y, ycov = util.propagate(fn, x, cov)
     np.testing.assert_allclose(y, [3, 5, 7])
-    np.testing.assert_allclose(ycov, [[4, 0.4, 0.8], [0.4, 8, 1.2], [0.8, 1.2, 12]])
+    np.testing.assert_allclose(
+        ycov, [[4, 0.4, 0.8], [0.4, 8, 1.2], [0.8, 1.2, 12]], rtol=1e-3
+    )
 
     with pytest.warns(np.VisibleDeprecationWarning):
         y, ycov = util.propagate(fn, [1], [[2]])
     np.testing.assert_allclose(y, 3)
-    np.testing.assert_allclose(ycov, 8)
+    np.testing.assert_allclose(ycov, 8, rtol=1e-3)
 
 
+@pytest.mark.skipif(not scipy_available, reason="needs scipy")
 def test_propagate_2():
     cov = [
         [1.0, 0.1, 0.2],
@@ -545,7 +560,7 @@ def test_propagate_2():
     with pytest.warns(np.VisibleDeprecationWarning):
         y, ycov = util.propagate(fn, x, cov)
     np.testing.assert_equal(y, fn(x))
-    np.testing.assert_allclose(ycov, np.einsum("ij,kl,jl", a, a, cov))
+    np.testing.assert_allclose(ycov, np.einsum("ij,kl,jl", a, a, cov), rtol=1e-3)
 
     def fn(x):
         return np.linalg.multi_dot([x.T, cov, x])
@@ -554,9 +569,10 @@ def test_propagate_2():
         y, ycov = util.propagate(fn, x, cov)
     np.testing.assert_equal(y, fn(np.array(x)))
     jac = 2 * np.dot(cov, x)
-    np.testing.assert_allclose(ycov, np.einsum("i,k,ik", jac, jac, cov))
+    np.testing.assert_allclose(ycov, np.einsum("i,k,ik", jac, jac, cov), rtol=1e-3)
 
 
+@pytest.mark.skipif(not scipy_available, reason="needs scipy")
 def test_propagate_3():
     # matrices with full zero rows and columns are supported
     cov = [
@@ -575,6 +591,7 @@ def test_propagate_3():
     np.testing.assert_allclose(ycov, [[4, 0.0, 0.8], [0.0, 0.0, 0.0], [0.8, 0.0, 12]])
 
 
+@pytest.mark.skipif(not scipy_available, reason="needs scipy")
 def test_propagate_on_bad_input():
     cov = [[np.nan, 0.0], [0.0, 1.0]]
     x = [1.0, 2.0]
@@ -596,102 +613,6 @@ def test_propagate_on_bad_input():
             util.propagate(fn, x, cov)
 
 
-def test_jacobi():
-    n = 100
-    x = np.linspace(-5, 5, n)
-    dx = np.abs(x) * 1e-3 + 1e-3
-    tol = 1e-3
-
-    y, jac = util._jacobi(np.exp, x, dx, tol)
-    np.testing.assert_equal(y, np.exp(x))
-    jac_ref = np.zeros((n, n))
-    for i in range(n):
-        jac_ref[i, i] = np.exp(x[i])
-    np.testing.assert_allclose(jac, jac_ref)
-
-    x = np.linspace(1e-10, 5, n)
-    dx = np.abs(x) * 1e-3
-    y, jac = util._jacobi(np.log, x, dx, tol)
-    jac_ref = np.zeros((n, n))
-    for i in range(n):
-        jac_ref[i, i] = 1 / x[i]
-    np.testing.assert_allclose(jac, jac_ref)
-
-
-def test_jacobi_on_bad_input():
-    x = np.array([1])
-    dx = np.array([0.1])
-    y, jac = util._jacobi(lambda x: np.nan, x, dx, 1e-3)
-
-    np.testing.assert_equal(y, np.nan)
-    np.testing.assert_equal(jac, np.nan)
-
-    x = np.array([1])
-    dx = np.array([0])
-    y, jac = util._jacobi(lambda x: x**2, x, dx, 1e-3)
-
-    np.testing.assert_equal(y, 1)
-    np.testing.assert_equal(jac, 0)
-
-    x = np.array([np.nan])
-    dx = np.array([0.1])
-    y, jac = util._jacobi(lambda x: x**2, x, dx, 1e-3)
-
-    np.testing.assert_equal(y, np.nan)
-    np.testing.assert_equal(jac, np.nan)
-
-    x = np.array([1])
-    dx = np.array([np.nan])
-    with pytest.raises(AssertionError):
-        # dx must be >= 0
-        util._jacobi(lambda x: x**2, x, dx, 1e-3)
-
-
-@pytest.mark.parametrize("fail", (False, True))
-def test_jacobi_low_resolution(fail, capsys):
-    x = np.array([1])
-    dx = np.array([1])
-
-    y, jac = util._jacobi(
-        lambda x: np.exp(x.astype(np.float32)),
-        x,
-        dx,
-        1e-10 if fail else 1e-3,
-        debug=True,
-    )
-
-    assert fail == ("no convergence" in capsys.readouterr()[0])
-
-    np.testing.assert_allclose(y, np.exp(1), rtol=1e-3)
-    np.testing.assert_allclose(jac, np.exp(1), rtol=1e-3)
-
-
-def test_jacobi_divergence_1(capsys):
-    rng = np.random.default_rng(1)
-
-    x = np.array([1])
-    dx = np.array([1])
-
-    y, jac = util._jacobi(lambda x: rng.normal(), x, dx, 0.1, debug=True)
-
-    assert "divergence" in capsys.readouterr()[0]
-
-    np.testing.assert_allclose(y, 0, atol=1)
-    np.testing.assert_equal(jac, np.nan)
-
-
-def test_jacobi_divergence_2(capsys):
-    x = np.array([1e-10])
-    dx = np.array([0.1])
-
-    y, jac = util._jacobi(lambda x: 1 / x, x, dx, 1e-3, debug=True)
-
-    assert "divergence" in capsys.readouterr()[0]
-
-    np.testing.assert_allclose(y, 1e10)
-    np.testing.assert_equal(jac, np.nan)
-
-
 def test_iterate():
     assert list(util._iterate(1)) == [1]
     assert list(util._iterate([1, 2])) == [1, 2]
@@ -703,14 +624,23 @@ def test_replace_none():
 
 
 def test_progressbar(capsys):
-    with util._ProgressBar(max_value=4) as bar:
+    with util.ProgressBar(max_value=4) as bar:
         for i in range(4):
             bar += 1
     stdout, stderr = capsys.readouterr()
     assert stdout == "\r0 %\r25 %\r50 %\r75 %\r100 %\r     "
 
 
-def test_progressbar_2(capsys):
+def test_progressbar_no_ipykernel(capsys):
+    with hide_modules("ipykernel"):
+        with util.ProgressBar(max_value=4) as bar:
+            for i in range(4):
+                bar += 1
+        stdout, stderr = capsys.readouterr()
+        assert stdout == "\r0 %\r25 %\r50 %\r75 %\r100 %\r     "
+
+
+def test_progressbar_html(capsys):
     import sys
 
     m_iostream = pytest.importorskip("ipykernel.iostream")
@@ -727,13 +657,14 @@ def test_progressbar_2(capsys):
         sys.stdout.write(msg._repr_html_())
 
     original_stdout = sys.stdout
+    # make ProgressBar think it is running in Jupyter
     sys.stdout = OutStream()
 
     # monkey-patching our mockups
     m_iostream.OutStream = OutStream
     m_display.display = display
 
-    with util._ProgressBar(max_value=4) as bar:
+    with util.ProgressBar(max_value=4) as bar:
         for i in range(4):
             bar += 1
 
@@ -780,5 +711,60 @@ def test_smart_sampling_1(fn_expected):
 
 
 def test_smart_sampling_2():
-    with pytest.warns(RuntimeWarning):
-        util._smart_sampling(np.log, 1e-10, 1, tol=1e-10)
+    # should not raise a warning
+    x, y = util._smart_sampling(np.log, 1e-10, 1, tol=1e-5)
+    assert 0 < len(x) < 1000
+
+
+def test_smart_sampling_3():
+    def step(x):
+        return np.where(x > 0.5, 0, 1)
+
+    with pytest.warns(RuntimeWarning, match="Iteration limit"):
+        x, y = util._smart_sampling(step, 0, 1, tol=0)
+    assert 0 < len(x) < 80
+
+
+def test_smart_sampling_4():
+    from time import sleep
+
+    def step(x):
+        sleep(0.1)
+        return np.where(x > 0.5, 0, 1)
+
+    with pytest.warns(RuntimeWarning, match="Time limit"):
+        x, y = util._smart_sampling(step, 0, 1, maxtime=0)
+    assert 0 < len(x) < 10
+
+
+@pytest.mark.parametrize(
+    "x,expected",
+    [([1, 2, 3], False), ([-1, 1000, 100000], False), ([1, 10, 100], True)],
+)
+def test_detect_log_spacing_1(x, expected):
+    assert util._detect_log_spacing(x) == expected
+
+
+def test_optional_module_for_1():
+    with optional_module_for("foo"):
+        import iminuit  # noqa
+
+
+def test_optional_module_for_2():
+    from iminuit.warnings import OptionalDependencyWarning
+
+    with pytest.warns(
+        OptionalDependencyWarning, match="foo requires optional package 'foobarbaz'"
+    ):
+        with optional_module_for("foo"):
+            import foobarbaz  # noqa
+
+
+def test_optional_module_for_3():
+    from iminuit.warnings import OptionalDependencyWarning
+
+    with pytest.warns(
+        OptionalDependencyWarning, match="foo requires optional package 'foo'"
+    ):
+        with optional_module_for("foo", replace={"foobarbaz": "foo"}):
+            import foobarbaz  # noqa
